"""esbmtk: A general purpose Earth Science box model toolkit Copyright
     (C), 2020-2021 Ulrich G. Wortmann

     This program is free software: you can redistribute it and/or
     modify it under the terms of the GNU General Public License as
     published by the Free Software Foundation, either version 3 of
     the License, or (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see
     <https://www.gnu.org/licenses/>.

"""
from __future__ import annotations
import typing as tp
from numba import njit
from esbmtk.utility_functions import (
    __checkkeys__,
    __addmissingdefaults__,
    __checktypes__,
    register_return_values,
)

if tp.TYPE_CHECKING:
    from esbmtk import Connection, Connect, Q_

# declare numpy types
# NDArrayFloat = npt.NDArray[np.float64]


# @njit(fastmath=True)
def weathering(pco2t, p) -> float | tuple:
    """Calculates weathering as a function pCO2 concentration

    :param pco2: float current pco2
    :param pco2_0: float reference pco2
    :area_fraction: float area/total area
    :param ex: exponent
    :f0: flux at pco2_0
    :returns:  F_w or F_w, F_w_i

    F_w = area_fraction * f0 * (pco2/pco2_0)**ex
    """
    pco2_0, area_fraction, ex, f0, isotopes = p
    if isotopes:
        pco2, pco2i = pco2t
        F_w = area_fraction * f0 * (pco2 / pco2_0) ** ex
        F_w_i = F_w * pco2i / pco2
        rv = F_w, F_w_i
    else:
        pco2 = pco2t
        F_w = area_fraction * f0 * (pco2 / pco2_0) ** ex
        rv = F_w

    return rv


def init_weathering(
    c: Connection | Connect,
    pco2: float,
    pco2_0: float | str | Q_,
    area_fraction: float,
    ex: float,
    f0: float | str | Q_,
):
    """Creates a new external code instance

    :param c: Connection
    :param pco2: float current pco2
    :param pco2_0: float reference pco2
    :area_fraction: float area/total area
    :param ex: exponent
    :f0: flux at pco2_0

    """
    from esbmtk import ExternalCode, check_for_quantity

    print(f" initializing weathering with isotopes = {c.sink.isotopes}")

    f0 = check_for_quantity(f0, "mol/year").magnitude
    pco2_0 = check_for_quantity(pco2_0, "ppm").magnitude
    p = (pco2_0, area_fraction, ex, f0, c.sink.isotopes)
    c.fh.ftype = "computed"
    ec = ExternalCode(
        name=f"ec_weathering_{c.id}",
        fname="weathering",
        ftype="std",
        species=c.sink.species,
        function_input_data=[pco2],
        function_params=p,
        register=c.model,
        return_values=[
            {f"F_{c.fh.full_name}": "fww"},
        ],
    )
    c.mo.lpc_f.append(ec.fname)

    return ec


def init_gas_exchange(c: Connection | Connect):
    """Create an ExternalCode instance for gas exchange reactions

    Parameters
    ----------
    c : Connection | Connect
        connection instance
    """
    from esbmtk import ExternalCode, check_for_quantity

    c.fh.ftype = "computed"
    swc = c.sink.register.swc  # sink - liquid
    piston_velocity = (
        check_for_quantity(c.piston_velocity, "m/yr").to("meter/year").magnitude
    )
    area = check_for_quantity(c.area, "m**2").to("meter**2").magnitude
    scale = area * piston_velocity
    p = (
        scale,
        c.water_vapor_pressure,
        c.solubility,
        swc.a_db,
        swc.a_dg,
        swc.a_u,
        c.source.isotopes,
    )

    ec = ExternalCode(
        name=f"gas_exchange{c.id}",
        fname="gas_exchange",
        ftype="std",
        species=c.source.species,
        function_input_data=[c.source, c.sink, c.ref_species],
        function_params=p,
        register=c.model,
        return_values=[
            {f"F_{c.fh.full_name}": "gex"},
        ],
    )
    c.mo.lpc_f.append(ec.fname)

    return ec


def gas_exchange(
    gas_c: float | tuple,
    liquid_c: float | tuple,
    gas_aq: float,
    p: tuple,
) -> float | tuple:
    """Calculate the gas exchange flux across the air sea interface
    for co2 including isotope effects.

    Parameters
    ----------
    gas_c : float | tuple
        gas concentration in atmosphere
    liquid_c : float | tuple
        reference species in liquid phase, e.g., DIC
    gas_aq: float
        dissolved gas concentration, e.g., CO2aq
    p : tuple
        parameters, see init_gas_exchange

    Returns
    -------
    float | tuple
        gas flux across the air/sea interface

    Note that the sink delta is co2aq as returned by the carbonate VR
    this equation is for mmol but esbmtk uses mol, so we need to
    multiply by 1E3

    The Total flux across interface dpends on the difference in either
    concentration or pressure the atmospheric pressure is known, as gas_c, and
    we can calculate the equilibrium pressure that corresponds to the dissolved
    gas in the water as [CO2]aq/beta.

    Conversely, we can convert the the pCO2 into the amount of dissolved CO2 =
    pCO2 * beta

    The h/c ratio in HCO3 estimated via h/c in DIC. Zeebe writes C12/C13 ratio
    but that does not work. the C13/C ratio results however in -8 permil
    offset, which is closer to observations
    """
    scale, p_H2O, solubility, a_db, a_dg, a_u, isotopes = p

    if isotopes:
        gas_c, gas_c_l = gas_c
        liquid_c, liquid_c_l = liquid_c

    # Solubility with correction for pH2O
    beta = solubility * (1 - p_H2O)
    # f as afunction of solubility difference
    f = scale * (beta * gas_c - gas_aq * 1e3)
    rv = f

    if isotopes:  # isotope ratio of DIC
        Rt = (liquid_c - liquid_c_l) / liquid_c
        # get heavy isotope concentrations in atmosphere
        gas_c_h = gas_c - gas_c_l  # gas heavy isotope concentration
        # get exchange of the heavy isotope
        f_h = scale * a_u * (a_dg * gas_c_h * beta - Rt * a_db * gas_aq * 1e3)
        f_l = f - f_h  # the corresponding flux of the light isotope
        rv = f, f_l

    return rv



# import classes from the esbmtk library
from esbmtk import (
    Model,  # the model class
    Reservoir,  # the reservoir class
    Connection,  # the connection class
    Source,  # the source class
    Sink,  # sink class
    Q_,  # Quantity operator
    register_user_function,
)

# define the basic model parameters
M = Model(
    stop="1000 yr",  # end time of model
    timestep="1 yr",  # upper limit of time step
    element=["Phosphor"],  # list of element definitions
    # parse_model=False,
)

# boundary conditions
F_w = M.set_flux("45 Gmol", "year", M.P)  # P @280 ppm (Filipelli 2002)
tau = Q_("100 year")  # PO4 residence time in surface box
R_e = 1 - 0.01  # About 1% of the exported P is buried in the deep ocean
thc = "20*Sv"  # Thermohaline circulation in Sverdrup

# Source definitions
Source(
    name="weathering",
    species=M.PO4,
    register=M,  # i.e., the instance will be available as M.weathering
)
Sink(
    name="burial",
    species=M.PO4,
    register=M,  #
)

# reservoir definitions
Reservoir(
    name="S_b",  # box name
    species=M.PO4,  # species in box
    register=M,  # this box will be available as M.S_b
    volume="3E16 m**3",  # surface box volume
    concentration="0 umol/l",  # initial concentration
)
Reservoir(
    name="D_b",  # box name
    species=M.PO4,  # species in box
    register=M,  # this box will be available M.D_b
    volume="100E16 m**3",  # deeb box volume
    concentration="0 umol/l",  # initial concentration
)

Connection(
    source=M.weathering,  # source of flux
    sink=M.S_b,  # target of flux
    rate=F_w,  # rate of flux
    id="river",  # connection id
    ctype="regular",
)

Connection(  # thermohaline downwelling
    source=M.S_b,  # source of flux
    sink=M.D_b,  # target of flux
    ctype="scale_with_concentration",
    scale=thc,
    id="downwelling_PO4",
    # ref_reservoirs=M.S_b, defaults to the source instance
)
Connection(  # thermohaline upwelling
    source=M.D_b,  # source of flux
    sink=M.S_b,  # target of flux
    ctype="scale_with_concentration",
    scale=thc,
    id="upwelling_PO4",
)

Connection(  #
    source=M.S_b,  # source of flux
    sink=M.D_b,  # target of flux
    ctype="scale_with_concentration",
    scale=M.S_b.volume / tau,
    id="primary_production",
)

# --- Adding a user supplied function into the model --- #

""" To add a user supplied function, we need a local file
(e.g., my_functions.py) where we define our new function
(e.g.,  my_burial()). In addition we need some code that
describes the relationship of this function with the model
(e.g., add_my_function()).

Assuming that the functions my_burial() and
add_my_burial() exist in the local file my_functions.py,
we can register the module my_functions.py, and the
function my_burial() with the ESBMTK model using the aptly
named register_user_function() function. Note that the
function name can also be a list of several function names.
"""
register_user_function(M, "my_functions", "my_burial")

""" Next we import the import the function add_my_burial()
into this script, so that we connect it with the model.
Unlike my_burial() the add_my_burial() function can also be
defined in this script.
Note that unless add_my_burial() contains code to correctly
interpret quantities, you need to make sure that the parameters
passed to add_my_burial() are float numbers in model units
(e.g., liter, year, kg) etc. That is why the below code
uses M.D_b.volume.magnitude rather than  M.D_b.volume (which
is a quantity)
"""
from my_functions import add_my_burial

# Last but not least, apply the new function
add_my_burial(
    M.D_b,  # Source
    M.burial,  # Sink
    M.PO4,  # Species
    M.D_b.volume.magnitude / 2000.0,  # Scale
)

M.run()
M.plot([M.S_b, M.D_b], fn="po4_1.png")
# M.save_state(directory="state_po41")



import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import pathlib as pl
from esbmtk import (
    Model,  # the model class
    Reservoir,
    ExternalData,
    DataField,
)

# define the basic model parameters
M = Model(
    stop="6 Myr",  # end time of model
    timestep="1 kyr",  # upper limit of time step
    element=["Carbon"],  # list of element definitions
    concentration_unit="mol/kg",
)

Reservoir(
    name="S_b",  # Name of reservoir group
    geometry=[-200, -800, 1],  # upper, lower, fraction
    concentration="1 mmol/kg",
    species=M.DIC,
    register=M,
)
print(f"M.S_b.area = {M.S_b.area:.2e}")
print(f"M.S_b.sed_area = {M.S_b.sed_area:.2e}")
print(f"M.S_b.volume = {M.S_b.volume:.2e}\n")

# return the ocean area at a given depth in m**2
print(f"M.hyp.area(0) = {M.hyp.area(0):.2e}")

# return the area between 2 depth datums in m**2
print(f"M.hyp.area_dz(0, -200) = {M.hyp.area_dz(0, -200):.2e}")

# return the volume between 2 depth datums in m**3
print(f"M.hyp.volume(0,-200) = {M.hyp.volume(0,-200):.2e}")

# return the total surface area of earth in m**3
print(f"M.hyp.sa = {M.hyp.sa:.2e}")

# get figure of data
fn: str = "Hypsometric_Curve_05m.csv"  # file name
cwd: pl.Path = pl.Path.cwd()  # get the current working directory
fqfn: pl.Path = pl.Path(f"{cwd}/{fn}")  # fully qualified file name

if not fqfn.exists():  # check if file exist
    raise FileNotFoundError(f"Cannot find file {fqfn}")

df: pd.DataFrame = pd.read_csv(fqfn)  # read csv data

x = df.iloc[:, 0]
y = df.iloc[:, 1]

elevation = np.arange(1000, -6001, -1)
plt.style.use(["ggplot"])
fig, ax = plt.subplots()
ax.plot(y, x, color="C1", label="Data", linewidth=3)  # create a line plot
ax.plot(
    M.hyp.hypdata, elevation, color="C0", label="Spline", linewidth=1
)  # create a line plot
ax.set_xlabel("Surface Area [%]")
ax.set_ylabel("Depth [m]")
ax.axis([1, 0, -6000, 1000])
ticks = ax.get_xticks() * 100
ticks = ticks.astype(int)
ax.xaxis.set_major_formatter("{x%.0f}")
ax.set_xticklabels(ticks)
plt.legend()
plt.savefig("hyp.png")
plt.show()  # display figure
